from ._internal import DEFAULT, defaultarguments, DoreahConfig
from .database import Database, Ref
from .pyhp2 import file as pyhpfile

import secrets
import time
from nimrodel import EAPI
from threading import Lock
from bottle import response as bottleresponse
from bottle import request, HTTPResponse, redirect
import pkg_resources

config = DoreahConfig("auth",
	multiuser=True,
	cookieprefix="doreahauth",
	defaultpw="admin"
)


db = Database(file="authdb.ddb")
cookie_token_name = config["cookieprefix"] + "_sessiontoken"
authapi = EAPI(path="auth",delay=True)

CHALLENGE_BITS = 32
TOKEN_BITS = 32


COMMON = 3590390094909444
GENERATOR = 2
MODULO = """
	FFFFFFFF FFFFFFFF C90FDAA2 2168C234 C4C6628B 80DC1CD1
	29024E08 8A67CC74 020BBEA6 3B139B22 514A0879 8E3404DD
	EF9519B3 CD3A431B 302B0A6D F25F1437 4FE1356D 6D51C245
	E485B576 625E7EC6 F44C42E9 A637ED6B 0BFF5CB6 F406B7ED
	EE386BFB 5A899FA5 AE9F2411 7C4B1FE6 49286651 ECE45B3D
	C2007CB8 A163BF05 98DA4836 1C55D39A 69163FA8 FD24CF5F
	83655D23 DCA3AD96 1C62F356 208552BB 9ED52907 7096966D
	670C354E 4ABC9804 F1746C08 CA18217C 32905E46 2E36CE3B
	E39E772C 180E8603 9B2783A2 EC07A28F B5C55DF0 6F4C52C9
	DE2BCBF6 95581718 3995497C EA956AE5 15D22618 98FA0510
	15728E5A 8AACAA68 FFFFFFFF FFFFFFFF
"""


CYCLE_CHALLENGE = 120
SESSION_EXPIRE = 3600000

### DERIVED

COMMON = hex(COMMON)
MODULO = int(MODULO.replace(" ","").replace("\n","").replace("\t",""),16)


challenge_current = None
challenge_old = None
stop_issue = 0
expire = 0
lock = Lock()
updatelock = Lock()







def combine(a,b):
	a = int(a,16)
	b = int(b,16)
	#return (a ** b) % P
	result = pow(a,b,MODULO)
	return hex(result)[2:]


class User(db.DBObject):
	__primary__ = "username",
	username: str
	pubkey: str


	def setpw(self,password):
		pw = str.encode(password).hex()
		pw1 = combine(hex(GENERATOR),pw)
		pub = combine(pw1,COMMON)
		self.pubkey = pub
		db.save()

	def validate_challenge(self,challenge,response):
		update()
		lock.acquire()
		if challenge not in [challenge_current, challenge_old]:
			lock.release()
			print("Invalid challenge")
			return False

		pub = str(self.pubkey)

		if combine(pub,challenge) == combine(response,COMMON):
			# invalidate challenges
			stop_issue = 0
			expire = 0
			update()

			lock.release()
			return True

		else:
			print("Invalid response")
			lock.release()
			return False

if not config["multiuser"]:
	defaultuser = User(username="")
	if defaultuser.pubkey == "": defaultuser.setpw(config["defaultpw"])


sessions = {} #token -> timestamp expire
class Session(db.DBObject):
	token: str
	timestamp: float
	user: User = Ref(User,exclusive=False)



def update():
	updatelock.acquire()
	global challenge_current, stop_issue, expire
	now = time.time()
	if now > expire:
		challenge_old = None
		challenge_current = generate_challenge()
		stop_issue = now + CYCLE_CHALLENGE
		expire = stop_issue + CYCLE_CHALLENGE
	elif now > stop_issue:
		challenge_old = challenge_current
		challenge_current = generate_challenge()
		stop_issue = expire #expire value of old challenge is stop issue value of new
		expire = stop_issue + CYCLE_CHALLENGE
	updatelock.release()
	db.save()

def generate_challenge():
	return secrets.token_hex(CHALLENGE_BITS)

def generate_token():
	return secrets.token_hex(TOKEN_BITS)

@authapi.get("challenge")
def get_challenge():
	global challenge_current, stop_issue, expire
	update()
	return {"challenge":challenge_current, "modulo":MODULO,"generator":GENERATOR}


@authapi.post("authenticate")
def get_token(user,challenge,response):
	if not config["multiuser"]:
		u = defaultuser
	else:
		u = db.getby(User,username=user)
	if u.validate_challenge(challenge,response):
		token = generate_token()
		#sessions[token] = time.time() + SESSION_EXPIRE
		Session(token=token,timestamp=time.time() + SESSION_EXPIRE,user=u)
		bottleresponse.set_header("Set-Cookie",cookie_token_name + "=" + token)# + "; HttpOnly")
		db.save()
		return {"token":token,"cookie_name":cookie_token_name}
	else:
		return {"error":"Access denied."}


def check(request):
	token = request.get_cookie(cookie_token_name)
	if token is None: return False
	for ses in db.getall(Session):
		if ses.token == token:
			session = ses
			break
	else:
		return False
	if session.timestamp < time.time(): return False

	session.timestamp = time.time() + SESSION_EXPIRE
	return True

def authenticated(func):

	def newfunc(*args,**kwargs):
		if check(request):
			return func(*args,**kwargs)
		else:
			redirect("/")

	return newfunc


def get_login_page(stylesheets=[]):
	return pyhpfile(pkg_resources.resource_filename(__name__,"res/login.pyhp"),{"get_challenge":get_challenge,"css":stylesheets})


db.save()



from ._internal import DoreahConfig

config = DoreahConfig("pyhp",
	interpret=str,
	version=1
)

if config["version"] == 1:
	print("Using PYHP version 1")
	from .pyhp1 import *
elif config["version"] == 2:
	print("Using PYHP version 2")
	from .pyhp2 import *



def _ANSI(code=0):
	return "\033[" + str(code) + "m"
def _ANSIRGB(r=0,g=0,b=0):
	return _ANSI("38;2;" + str(r) + ";" + str(g) + ";" + str(b))
def _ANSIHEX(code="#000000"):
	if code[0] == "#": code = code[1:]
	r,g,b = int(code[0:2],16),int(code[2:4],16),int(code[4:6],16)
	return _ANSIRGB(r=r,g=g,b=b)

def _ANSICOLOR(code):
	try:
		if isinstance(code,str):
			if code.lower() in colornames:
				code = colornames[code.lower()]
			return _ANSIHEX(code), _ANSI()
		elif isinstance(code,tuple):
			return _ANSIRGB(*code), _ANSI()
		else:
			return "",""
	except:
		return "",""

colornames = {
	"aliceblue": "#f0f8ff",
	"antiquewhite": "#faebd7",
	"aqua": "#00ffff",
	"aquamarine": "#7fffd4",
	"azure": "#f0ffff",
	"beige": "#f5f5dc",
	"bisque": "#ffe4c4",
	"black": "#000000",
	"blanchedalmond": "#ffebcd",
	"blue": "#0000ff",
	"blueviolet": "#8a2be2",
	"brown": "#a52a2a",
	"burlywood": "#deb887",
	"cadetblue": "#5f9ea0",
	"chartreuse": "#7fff00",
	"chocolate": "#d2691e",
	"coral": "#ff7f50",
	"cornflowerblue": "#6495ed",
	"cornsilk": "#fff8dc",
	"crimson": "#dc143c",
	"cyan": "#00ffff",
	"darkblue": "#00008b",
	"darkcyan": "#008b8b",
	"darkgoldenrod": "#b8860b",
	"darkgray": "#a9a9a9",
	"darkgreen": "#006400",
	"darkgrey": "#a9a9a9",
	"darkkhaki": "#bdb76b",
	"darkmagenta": "#8b008b",
	"darkolivegreen": "#556b2f",
	"darkorange": "#ff8c00",
	"darkorchid": "#9932cc",
	"darkred": "#8b0000",
	"darksalmon": "#e9967a",
	"darkseagreen": "#8fbc8f",
	"darkslateblue": "#483d8b",
	"darkslategray": "#2f4f4f",
	"darkslategrey": "#2f4f4f",
	"darkturquoise": "#00ced1",
	"darkviolet": "#9400d3",
	"deeppink": "#ff1493",
	"deepskyblue": "#00bfff",
	"dimgray": "#696969",
	"dimgrey": "#696969",
	"dodgerblue": "#1e90ff",
	"firebrick": "#b22222",
	"floralwhite": "#fffaf0",
	"forestgreen": "#228b22",
	"fuchsia": "#ff00ff",
	"gainsboro": "#dcdcdc",
	"ghostwhite": "#f8f8ff",
	"goldenrod": "#daa520",
	"gold": "#ffd700",
	"gray": "#808080",
	"green": "#008000",
	"greenyellow": "#adff2f",
	"grey": "#808080",
	"honeydew": "#f0fff0",
	"hotpink": "#ff69b4",
	"indianred": "#cd5c5c",
	"indigo": "#4b0082",
	"ivory": "#fffff0",
	"khaki": "#f0e68c",
	"lavenderblush": "#fff0f5",
	"lavender": "#e6e6fa",
	"lawngreen": "#7cfc00",
	"lemonchiffon": "#fffacd",
	"lightblue": "#add8e6",
	"lightcoral": "#f08080",
	"lightcyan": "#e0ffff",
	"lightgoldenrodyellow": "#fafad2",
	"lightgray": "#d3d3d3",
	"lightgreen": "#90ee90",
	"lightgrey": "#d3d3d3",
	"lightpink": "#ffb6c1",
	"lightsalmon": "#ffa07a",
	"lightseagreen": "#20b2aa",
	"lightskyblue": "#87cefa",
	"lightslategray": "#778899",
	"lightslategrey": "#778899",
	"lightsteelblue": "#b0c4de",
	"lightyellow": "#ffffe0",
	"lime": "#00ff00",
	"limegreen": "#32cd32",
	"linen": "#faf0e6",
	"magenta": "#ff00ff",
	"maroon": "#800000",
	"mediumaquamarine": "#66cdaa",
	"mediumblue": "#0000cd",
	"mediumorchid": "#ba55d3",
	"mediumpurple": "#9370db",
	"mediumseagreen": "#3cb371",
	"mediumslateblue": "#7b68ee",
	"mediumspringgreen": "#00fa9a",
	"mediumturquoise": "#48d1cc",
	"mediumvioletred": "#c71585",
	"midnightblue": "#191970",
	"mintcream": "#f5fffa",
	"mistyrose": "#ffe4e1",
	"moccasin": "#ffe4b5",
	"navajowhite": "#ffdead",
	"navy": "#000080",
	"oldlace": "#fdf5e6",
	"olive": "#808000",
	"olivedrab": "#6b8e23",
	"orange": "#ffa500",
	"orangered": "#ff4500",
	"orchid": "#da70d6",
	"palegoldenrod": "#eee8aa",
	"palegreen": "#98fb98",
	"paleturquoise": "#afeeee",
	"palevioletred": "#db7093",
	"papayawhip": "#ffefd5",
	"peachpuff": "#ffdab9",
	"peru": "#cd853f",
	"pink": "#ffc0cb",
	"plum": "#dda0dd",
	"powderblue": "#b0e0e6",
	"purple": "#800080",
	"rebeccapurple": "#663399",
	"red": "#ff0000",
	"rosybrown": "#bc8f8f",
	"royalblue": "#4169e1",
	"saddlebrown": "#8b4513",
	"salmon": "#fa8072",
	"sandybrown": "#f4a460",
	"seagreen": "#2e8b57",
	"seashell": "#fff5ee",
	"sienna": "#a0522d",
	"silver": "#c0c0c0",
	"skyblue": "#87ceeb",
	"slateblue": "#6a5acd",
	"slategray": "#708090",
	"slategrey": "#708090",
	"snow": "#fffafa",
	"springgreen": "#00ff7f",
	"steelblue": "#4682b4",
	"tan": "#d2b48c",
	"teal": "#008080",
	"thistle": "#d8bfd8",
	"tomato": "#ff6347",
	"turquoise": "#40e0d0",
	"violet": "#ee82ee",
	"wheat": "#f5deb3",
	"white": "#ffffff",
	"whitesmoke": "#f5f5f5",
	"yellow": "#ffff00",
	"yellowgreen": "#9acd32"
}



name = "doreah"
version = 1,2,7
versionstr = ".".join(str(n) for n in version)
author = {
	"name": "Johannes Krattenmacher",
	"email": "python@krateng.dev",
	"github": "krateng"
}
requires = [
	"lxml",
	"mechanicalsoup",
	"requests",
	"pyyaml",
	"beautifulsoup4"
]
resources = [
	'res/login.pyhp'
]


__all__ = [
	"auth",
	"caching",
	"database",
	"io",
	"logging",
	"persistence",
	"pyhp",
	"regular",
	"scraping"
	"settings",
	"timing",
	"tsv"

]




# useful things for everyone


def deprecated(func):
	"""Function decorator to deprecate a function"""

	def newfunc(*args,**kwargs):
		print("\033[93m" + "Function " + func.__name__ + " is deprecated!" + "\033[0m")
		return func(*args,**kwargs)

	return newfunc
